"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var built_ins_1 = require("./built-ins");
var global_state_1 = require("./global-state");
var reaction_1 = require("./reaction");
var utils_1 = require("./utils");
var MAX_RUN_COUNT = 1000;
/**
 * ========================
 * 函数
 * ========================
 */
/**
 * 获取可观察的对象
 */
function observableObject(obj) {
    if (obj === void 0) { obj = {}; }
    if (utils_1.isPrimitive(obj)) {
        throw TypeError(obj + " \u662F\u57FA\u672C\u7C7B\u578B\uFF0Cdob \u4EC5\u652F\u6301\u975E\u57FA\u672C\u7C7B\u578B");
    }
    if (global_state_1.globalState.proxies.has(obj)) {
        return global_state_1.globalState.proxies.get(obj);
    }
    // proxy 惰性封装
    return toObservable(obj);
}
/**
 * 将 class 改造为可观察对象
 */
function observableObjectDecorator(target) {
    function wrap() {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return observableObject(new (target.bind.apply(target, [void 0].concat(args)))());
    }
    return wrap;
}
/**
 * 生成可观察的对象
 */
function toObservable(obj) {
    if (Object.getOwnPropertySymbols(obj).indexOf(global_state_1.globalState.ignoreDynamicSymbol) > -1) {
        // 如果对象忽略了动态化，直接返回
        return obj;
    }
    var dynamicObject;
    var builtIn = built_ins_1.default.get(obj.constructor);
    if (typeof builtIn === "function" || typeof builtIn === "object") {
        // 处理 map weakMap set weakSet
        dynamicObject = builtIn(obj, bindCurrentReaction, queueRunReactions, proxyValue);
    }
    else if (!builtIn) {
        dynamicObject = new Proxy(obj, {
            get: function (target, key, receiver) {
                var value = Reflect.get(target, key, receiver);
                // 如果 key 是 $raw，或者在 Action 中，直接返回原始对象
                if (key === "$raw") {
                    return target;
                }
                // 将子元素注册到 immutable，子元素可以找到其根节点对象，以及路径
                global_state_1.globalState.event.emit("get", { target: target, key: key, value: value });
                bindCurrentReaction(target, key);
                value = proxyValue(target, key, value);
                return value;
            },
            set: function (target, key, value, receiver) {
                // 旧值
                var oldValue = Reflect.get(target, key, receiver);
                // 如果新值是对象，优先取原始对象
                if (typeof value === "object" && value) {
                    value = value.$raw || value;
                }
                global_state_1.globalState.event.emit("set", { target: target, key: key, value: value, oldValue: oldValue });
                var result = Reflect.set(target, key, value, receiver);
                // 如果改动了 length 属性，或者新值与旧值不同，触发可观察队列任务
                // 这一步要在 Reflect.set 之后，确保触发时使用的是新值
                if (key === "length" || value !== oldValue) {
                    queueRunReactions(target, key);
                }
                return result;
            },
            deleteProperty: function (target, key) {
                var hasKey = Reflect.has(target, key);
                global_state_1.globalState.event.emit("deleteProperty", { target: target, key: key });
                var result = Reflect.deleteProperty(target, key);
                if (hasKey) {
                    queueRunReactions(target, key);
                }
                return result;
            }
        });
    }
    else {
        dynamicObject = obj;
    }
    global_state_1.globalState.proxies.set(obj, dynamicObject);
    global_state_1.globalState.proxies.set(dynamicObject, dynamicObject);
    global_state_1.globalState.originObjects.set(dynamicObject, obj);
    global_state_1.globalState.objectReactionBindings.set(obj, new Map());
    return dynamicObject;
}
/**
 * 返回 get 获取的结果，如果已有 proxy 就使用 proxy 返回，否则 toObservable 递归
 */
function proxyValue(target, key, value) {
    // 如果取值是 HTMLElement 对象，直接返回原对象，因为原生对象不能被封装
    if (typeof window !== "undefined" && value instanceof HTMLElement) {
        return value;
    }
    // 如果取的值是对象，优先取代理对象
    var resultIsObject = typeof value === "object" && value;
    var existProxy = resultIsObject && global_state_1.globalState.proxies.get(value);
    if (resultIsObject) {
        return existProxy || toObservable(value);
    }
    return value;
}
/**
 * 队列执行属于 target 对象、key 字段绑定的 observer 函数
 * 队列执行的意思是，执行，但不一定立即执行，比如包在 runInAction 中函数触发时，就不会立刻执行
 * 而是在整个函数体执行完毕后，收集完了队列再统一执行一遍
 */
function queueRunReactions(target, key) {
    // 如果处于严格模式，并且不在 batch 中，报错
    if (global_state_1.globalState.strictMode && global_state_1.globalState.batchDeep === 0) {
        throw Error("You are not allowed to modify observable value out of Action.");
    }
    var keyBinder = utils_1.getBinder(target, key).keyBinder;
    Array.from(keyBinder).forEach(function (reaction) {
        if (utils_1.inAction()) {
            // 在 Action 中，直接加入队列
            global_state_1.globalState.pendingReactions.add(reaction);
        }
        else {
            // 不在 Action 中，如果队列已经有值了，添加到队列并执行队列；如果队列无值，则直接执行
            if (global_state_1.globalState.pendingReactions.size === 0) {
                runReactionAsync(reaction);
            }
            else {
                global_state_1.globalState.pendingReactions.add(reaction);
                runPendingReactions();
            }
        }
    });
}
/**
 * 执行 observer
 */
function runReactionAsync(reaction) {
    reaction.run();
}
exports.runReactionAsync = runReactionAsync;
/**
 * 执行待执行的 observer 队列
 * 如果在 action 中，继续添加到 actionQueuedObservers
 * 如果不在 action 中，添加一个直接执行
 */
function runPendingReactions() {
    // 队列执行次数
    var currentRunCount = 0;
    global_state_1.globalState.pendingReactions.forEach(function (observer) {
        currentRunCount++;
        if (currentRunCount >= MAX_RUN_COUNT) {
            global_state_1.globalState.pendingReactions.clear();
            return;
        }
        runReactionAsync(observer);
    });
    // 清空执行 observe 队列
    global_state_1.globalState.pendingReactions.clear();
}
/**
 * 绑定当前 reaction
 */
function bindCurrentReaction(object, key) {
    // 将监听添加到这个 key 上，必须不在 runInAction 中才会跟踪
    // inBatch 只要非 0，说明 runInAction 结束了
    if (!global_state_1.globalState.currentReaction || utils_1.inAction()) {
        return;
    }
    var keyBinder = utils_1.getBinder(object, key).keyBinder;
    // 如果这个 key 还没有与当前 reaction 绑定，则绑定
    if (!keyBinder.has(global_state_1.globalState.currentReaction)) {
        keyBinder.add(global_state_1.globalState.currentReaction);
        global_state_1.globalState.currentReaction.addBinder(keyBinder);
    }
}
/**
 * 是否是可观察对象
 */
function isObservable(obj) {
    return (global_state_1.globalState.proxies.get(obj) === obj);
}
exports.isObservable = isObservable;
/**
 * 利用 reaction 做的快捷监听，callback 元素会被追踪，绑定的变量改动时，触发的依然是此 callback
 */
function observe(callback, delay) {
    var reaction = new reaction_1.Reaction("observe", function () {
        reaction.track(callback);
    }, delay);
    // 初始化就执行
    if (utils_1.inAction()) {
        // 如果在 action 中，直接添加到队列，等 action 执行完后，会自动执行此队列的
        global_state_1.globalState.pendingReactions.add(reaction);
    }
    else {
        reaction.run();
    }
    return {
        unobserve: function () {
            reaction.dispose();
        }
    };
}
exports.observe = observe;
/**
 * Enter next batch.
 */
function startBatch() {
    if (global_state_1.globalState.batchDeep === 0) {
        // 如果正在从 0 开始新的队列，清空原有队列
        global_state_1.globalState.pendingReactions = new Set();
    }
    global_state_1.globalState.batchDeep++;
    global_state_1.globalState.event.emit("startBatch", null);
}
exports.startBatch = startBatch;
/**
 * Exit the current batch.
 */
function endBatch() {
    if (--global_state_1.globalState.batchDeep === 0) {
        runPendingReactions();
    }
    global_state_1.globalState.event.emit("endBatch", null);
}
exports.endBatch = endBatch;
function Action(arg1, arg2, arg3) {
    if (arg2 === undefined) {
        return runInAction.call(this, arg1, arg1.name);
    }
    return actionDecorator.call(this, arg1, arg1.constructor.name + "." + arg2, arg3);
}
exports.Action = Action;
/**
 * Action 装饰器，自带 runInAction 效果
 */
function actionDecorator(target, propertyKey, descriptor) {
    var func = descriptor.value;
    return {
        get: function () {
            var _this = this;
            return function () {
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                return runInAction(func.bind.apply(func, [_this].concat(args)), propertyKey);
            };
        },
        set: function (newValue) {
            return newValue;
        }
    };
}
/**
 * 在这里执行的方法，会在执行完后统一执行 observe
 * 注意：如果不用此方法包裹，同步执行代码会触发等同次数的 observe，而不会自动合并!
 * 同时，在此方法中使用到的变量不会触发依赖追踪！
 */
function runInAction(fn, debugName) {
    global_state_1.globalState.event.emit("runInAction", { debugName: debugName });
    startBatch();
    try {
        return fn();
    }
    finally {
        endBatch();
    }
}
// ==============================================
// Action [end]
// ==============================================
function observable(target) {
    if (target === void 0) { target = {}; }
    if (typeof target === "function") {
        return observableObjectDecorator(target);
    }
    else {
        return observableObject(target);
    }
}
exports.observable = observable;
/**
 * Static，使装饰的对象不会监听
 */
function Static(obj) {
    Object.defineProperty(obj, global_state_1.globalState.ignoreDynamicSymbol, {
        value: true
    });
    return obj;
}
exports.Static = Static;
//# sourceMappingURL=observer.js.map